#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ward']

package_data = \
{'': ['*'], 'ward': ['tests/*']}

install_requires = \
['colorama>=0.3.3',
 'termcolor>=1.1.0',
 'click>=7.0',
 'toml>=0.9.4',
 'pygments>=2.4.2',
 'pprintpp>=0.4.0']

extras_require = \
{":python_version < '3.7'": ['dataclasses>=0.1'],
 'dev': ['black>=19.10b0', 'flake8>=3.7.9', 'flit>=2.2.0', 'pycleanup>=0.0.1']}

entry_points = \
{'console_scripts': ['ward = ward.run:run']}

setup(name='ward',
      version='0.40.1b0',
      description='A modern Python 3 test framework for finding and fixing flaws faster.',
      author='Darren Burns',
      author_email='darrenb900@gmail.com',
      url='https://github.com/darrenburns/ward',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
